/*  Дерево эпизодической памяти.
Только для записи моторных правил.
Попытки совмещать запись моторных и ментальных оказываются очень громозкими и неестественными.
Стало ясно, что моторная эпиз.память формируется в районе гиппокампа, а ментальная - где-то в лобных долях (там тоже есть циклы).
К дереву прилагается массив для фиксации исторической последовательности эпизодов.

Дерево легко использовать для
1) предказаний, что будет, если после данного стимула применить такой-то ответ
2) выбора ответов с позитивным эффектом

В дереве 5 узлов, до узла Action включительно:
NodeAID|NodeSID|NodePID|Trigger|Action|PARAMS1|PARAMS2|...|PARAMSnnn
т.е. после Action идут его дочки типа PARAMS (числом от ни одной до сколько угодно)

значения PARAMS[] - это свойства эпизода: Effect|LifeTime
таким образом, к последнему узлу Action может прикрепплять неограниченное число PARAMS[].
Массив PARAMS[][] возвращается func getEpisodicPARAMS(nodeID int)[]int{
Только узел ActionID имеет PARAMS[][] !!!!

В файл пишется:
ID|ParentID|NodeAID|NodeSID|NodePID|Trigger|Action|Effect|LifeTime
Нет смысла выносить Effect|LifeTime|ParentID в отдельныю структуру со связью по ID
т.к. память это не съэкономит, но приведет к лишнему геммору.

Дерево не активируется по каким-то стимулам, в него просто делается запись Правил в виде PARAMS
и используется для получения информации.




ПОСЛЕДСТВИЯ:
Становятся ненужными:
type rules struct { - теперь любые Правила и их последовательности находятся быстро и в нужном контексте.
type TriggerAndAction struct { - это оказалась особенно вредная структура, требующая постоянно работу с ее картой для вытаскивания Правил и здорово усложняющая выборку.

Это повлечет за собой переделку и ментальных правил, которые оказываются просто частью эпиз.памяти.
Так же станет необходимой переделка огромного количества функций с участием Правил и эпиз.памяти.


Пример записи дерева в файл:
1|0|67|0|0|0|0|
2|1|67|38|0|0|0|
3|2|67|38|107|0|0|
4|3|67|38|107|1|0|
5|4|67|38|107|1|2|1|36987763|
6|4|67|38|107|1|3|3|36987843|-1|36987853|3|36987862|
7|1|67|39|0|0|0|
8|7|67|39|107|0|0|
9|8|67|39|107|1|0|
10|9|67|39|107|1|3|3|36987870|
11|7|67|39|100|0|0|
12|11|67|39|100|1|0|
13|12|67|39|100|1|3|3|36987879|
14|0|200|0|0|0|0|
15|14|200|39|0|0|0|
16|15|200|39|100|0|0|
17|16|200|39|100|1|0|
18|17|200|39|100|1|3|3|36987887|
 */
/////////////////////////////////////////////////////////////////////


package psychic

import (
	"BOT/lib"
	"strconv"
	"strings"
)

///////////////////////////////////////////

// не использовать EpisodicTree: notUsingEpisodicTree = true
var notUsingEpisodicTree = false
///////////////////////////////////////////


type EpisodicTreeNode struct { // узел дерева Эпизодической памяти
	ID int
	// Узлы дерева:
	NodeAID	int // конечный узел активной ветки дерева моторных автоматизмов
	NodeSID	int // конечный узел активной ветки дерева ситуации
	NodePID	int // субъективный узел активной ветки дерева проблем (последний активировавшийся)
	Trigger int// Cтимул (действий оператора)
	Action int // образ ответных действий - всегда ActionsImage

	// Свойства данного узла - массив PARAMS:  (могут быть как объективными, так и субъективными!)
	PARAMS [][]int
	/* [2] иммет два значения:
	Effect int // эффект от действий: -1 или 0 или 1
	LifeTime int // время появления узла в пульсах от рождения
	 */

	//Связь узлов дерева:
	Children []EpisodicTreeNode // дочерние узлы (ветвление) НЕ АДРЕСА, А РЕАЛЬНЫЕ ОБЪЕКТЫ
	ParentID int     // ID родителя
	ParentNode *EpisodicTreeNode  // адрес родителя

}
//////////////////////////////////////////////////////
var EpisodicTree EpisodicTreeNode
var EpisodicTreeNodeFromID=make(map[int]*EpisodicTreeNode)
var MapGwardEpisodicTreeNodeFromID=lib.RegNewMapGuard()

/*Связь узлов историческая - ID конечных узлов веток EpisodicTree подряд по порядку возникновения новых.
Записывается в файл "/memory_psy/episodic_history.txt" - сплошная строка EpisodicTreeNode.ID с разделителем "|"
В случе добавления newPARAMS в существующий EpisodicTreeNode пишется его EpisodicTreeNode.ID
так что в EpisodicHistoryArr могут появляться несколько EpisodicTreeNode.ID.

Поиск в EpisodicHistoryArr нужного EpisodicTreeNode.ID
При оптимизации во сне порядок сохраняется, но могут вырезаться отдельные составляющие.
При вырезании из EpisodicHistoryArr пишется в файл episodic_history.txt новая строка
и т.о. EpisodicHistoryArr []int остается с непрерывной цепочкой индексов.

Найти в массиве EpisodicHistoryArr[] индексы всех членов по значению:
iArr:=lib.FindIndexes(EpisodicHistoryArr,ID)
 */
var EpisodicHistoryArr []int

///////////////////////////////////////////////////////

/* Элемент отдельного Правила
для вывода целевой цепочки Правил и т.п.
 */
type Rule struct { //
	Trigger int// Cтимул (действий оператора)
	Action int // образ ответных действий - всегда ActionsImage
	Effect int // эффект от действий: -1 или 0 или 1
}
///////////////////////////////////////////////////////



//////////// ЗАПИСАТЬ В ДЕРЕВО НОВЫЙ ЭПИЗОД
func saveNewEpisodic(NodeAID int,NodeSID int,NodePID int,Trigger int,Action int,Effect int){

	if notUsingEpisodicTree{
		return
	}

//	getEpisodesArrFromConditions(0)

	if Trigger==0 || Action==0{ // так не должно быть! т.к. в Action заполнен массив PARAMS
		return
	}
	newPARAMS:= []int{Effect,LifeTime}
	var condArr = []int{NodeAID,NodeSID,NodePID,Trigger,Action}

	//проверяются дубликаты
	idOld, nodeOld := checkEpisodicBranchFromCondition(NodeAID,NodeSID,NodePID,Trigger,Action)
	if idOld > 0 {
		EpisodicHistoryArr=append(EpisodicHistoryArr,idOld)
		// проверить для существующего узла, есть ли в его  nodeOld.PARAMS новый newPARAMS
		params:=nodeOld.PARAMS
		for i := 0; i < len(params); i++ {
			if lib.EqualArrs(params[i], newPARAMS){
				EpisodicHistoryArr=append(EpisodicHistoryArr,idOld)
				return // есть уже newPARAMS, не нужно ничего делать
			}
		}
	}

	//lastNodeID:=addEpisodicFromNodeIDsToBrange(0,0,condArr,newPARAMS)
	lastNodeID:=addEpisodicFromNodeIDsToBrange(0,0,condArr,newPARAMS)
	EpisodicHistoryArr=append(EpisodicHistoryArr,lastNodeID)
	return
}
///////////////////////////////////////////////////////////////////////////////////





/////////////////////////////////////////// функции поддержки
/* Для создания нового узла ветки дерева
Возвращает ID узла и указатель на него - в любом случае, создан ли новый узел или найден такой же уже имеющийся.
*/
var lastIDNodeTree=0 // счетчик ID узлов, хранит ID хранит ID последнего созданного узла
func createEpisodicNodeTree(parent *EpisodicTreeNode,id int,NodeAID int,NodeSID int,NodePID int,Trigger int,Action int,PARAMS [][]int) (int,*EpisodicTreeNode) {
	if parent == nil{
		return 0,nil
	}

	if id==0{
		lastIDNodeTree++
		id=lastIDNodeTree
	}else{
		//		newW.ID=id
		if lastIDNodeTree<id{
			lastIDNodeTree=id
		}
	}

	var node EpisodicTreeNode
	node.ID = id
	node.ParentNode=parent
	node.ParentID=parent.ID
	node.NodeAID=NodeAID
	node.NodeSID=NodeSID
	node.NodePID=NodePID
	node.Trigger=Trigger
	node.Action=Action

	//
	node.PARAMS=PARAMS

	parent.Children = append(parent.Children, node)
	// четко находим новый вставленный член (а не &parent.Children[count-1])
	var newN *EpisodicTreeNode
	for i := 0; i < len(parent.Children); i++ {
		if parent.Children[i].ID == node.ID {
			newN = &parent.Children[i]
		}
	}

	lib.MapCheckBlock(MapGwardEpisodicTreeNodeFromID)
	EpisodicTreeNodeFromID[node.ID]=&node
	lib.MapFree(MapGwardEpisodicTreeNodeFromID)

	// т.к. append меняет длину массива, перетусовывая адреса, то нужно обновить адреса в AutomatizmTreeFromID:
	updatingEpisodicTreeNodeFromID(parent)// здесь потому, что при загрузке из файла нужно на лету получать адреса

	return id,newN
}

/////////////////////////////////////////////////
// добавить новую тройку параметров к уже имеющимся
/*
func addNewPARAM(pars [][]int,newPar []int)([][]int){
	var pOut [][]int
	if pars!=nil && len(pars)>0{
		for i := 0; i < len(pars); i++ {
			pOut=append(pOut,pars[i])
		}
	}
	pOut=append(pOut,newPar)

	return pOut
}*/
//////////////////////////////////
func checkEpisodicBranchFromCondition(NodeAID int,NodeSID int,NodePID int,Trigger int,Action int)(int,*EpisodicTreeNode){
	cond:= []int{NodeAID,NodeSID,NodePID,Trigger,Action,}
	//maxLev:=5
	id,level:=findEpisodicBrange(0, cond, &EpisodicTree)
	if id>0{
		// найти уровень lev0 ненулевых параметров
		lev0:=0
		for i := 0; i < len(cond); i++ {
			if cond[i]==0{
				break
			}
			lev0++
		}
		if lev0 > level{
			return 0,nil
		}
		//if maxLevel==maxLev{// есть такая ветка, просто возвращаем ее значения
			lib.MapCheckBlock(MapGwardEpisodicTreeNodeFromID)
			node,ok:=EpisodicTreeNodeFromID[id]
			lib.MapFree(MapGwardEpisodicTreeNodeFromID)
			if ok {
				return id, node
			}else{// такого не должно быть
				lib.TodoPanic("В func FindEpisodicTreeNodeFromCondition должно быть значение карты EpisodicTreeNodeFromID.") //вызвать панику
			}

		//}else{// нужно дорастить ветку
		//	return 0,nil
		//}
	}else{// ID==0 вообще нет совпадений, нужно наращивать с основы
		return 0,nil
	}

	return 0,nil
}
//////////
// рекурсивно корректируем адреса всех узлов
func updatingEpisodicTreeNodeFromID(rt *EpisodicTreeNode){
	if rt.ID>0 {
		lib.MapCheckBlock(MapGwardEpisodicTreeNodeFromID)
		rt.ParentNode=EpisodicTreeNodeFromID[rt.ParentID] // wr.ParentNode адрес меняется из=за corretsParent(,
		lib.MapFree(MapGwardEpisodicTreeNodeFromID)
		EpisodicTreeNodeFromID[rt.ID] = rt
	}
	if rt.Children == nil{// конец ветки
		return
	}
	for i := 0; i < len(rt.Children); i++ {
		updatingEpisodicTreeNodeFromID(&rt.Children[i])
	}
}
///////////////////////////////////////////////////////////





////////////////////////////////////////////////
// получить массив массивов PARAMS[3] последнего узла дерева
// Только узел c ActionID имеет PARAMS[][]
func getEpisodicPARAMS(nodeID int)[][]int{

	lib.MapCheckBlock(MapGwardEpisodicTreeNodeFromID)
	var actionNode,ok=EpisodicTreeNodeFromID[nodeID]
	lib.MapFree(MapGwardEpisodicTreeNodeFromID)
	if !ok{
		return nil
	}

	if actionNode.Action==0{
		return nil
	}

	return actionNode.PARAMS //[]int{}
}
////

/////////////////////////////////////////////////////////////////////////








/* Доращивание ветки, начиная с заданного узла fromID по массиву всех значений для ветки
newPARAMS - добавить к ветке, если она существует новый массив Effect|LifeTime
Возвращает ID конечного узла ветки.
*/
func addEpisodicFromNodeIDsToBrange(fromID int,lastLevel int,condArr []int,newPARAMS []int)int{

	var bNode *EpisodicTreeNode
	if fromID >0 {
		lib.MapCheckBlock(MapGwardEpisodicTreeNodeFromID)
		node, ok := EpisodicTreeNodeFromID[fromID]
		lib.MapFree(MapGwardEpisodicTreeNodeFromID)
		if !ok {
			return 0
		}
		bNode=node
	}else{ // не определен начальный узел, хотя он может и быть
		bNode=&EpisodicTree
	}

	lastNodeID,_:=addEpisodicNodesToBrange(bNode,lastLevel,condArr,newPARAMS)

	return lastNodeID
}
/* рекурсивно создать все недостающие уровни ветки по значениям condArr []int

*/
func addEpisodicNodesToBrange(fromNode *EpisodicTreeNode,level int,condArr []int,newPARAMS []int)(int,*EpisodicTreeNode){
	if fromNode==nil {
		return 0,nil
	}
	if level>=len(condArr) {
		return fromNode.ID,fromNode
	}
	//vArr := make([]int, len(condArr))// все vArr[n] имеют нулевые значения
	vArr :=[]int{0,0,0,0,0,}
	for i := 0; i <= level; i++ {// заполнить vArr до текущего уровня включительно, остальные - оставить нулями
		vArr[i] = condArr[i]
	}
	// с каждым уровнем vArr[n] добавляется новое значение для создания узла ветки следующего уровня.
// если такой узел есть (checkUnicum==true), то просто вернет его node
var node *EpisodicTreeNode
var pars [][]int

	//проверяются дубликаты потому, что новый узел нужно делать только если такого еще нет!!!
	idOld, nodeOld := checkEpisodicBranchFromCondition(vArr[0], vArr[1], vArr[2], vArr[3], vArr[4])
	if idOld > 0 {
		node = nodeOld
		// нарастить PARAMS
		if level==4 {
			pars = append(node.PARAMS, newPARAMS)
			node.PARAMS=pars
		}
		//node.PARAMS=append(node.PARAMS,pars)
	}else {
		_, node = createEpisodicNodeTree(fromNode, 0, vArr[0], vArr[1], vArr[2], vArr[3], vArr[4], pars)
		if level==4 {
			pars = append(node.PARAMS, newPARAMS)
			node.PARAMS=pars
		}
	}

	level++
	if level>=len(condArr) {
		return node.ID,node
	}
	id,node:= addEpisodicNodesToBrange(node,level, condArr, newPARAMS)
	return id,node
}
///////////////////////////////////////



/*  поиск конечного узла ветки (lastBrangeID) дерева (root - ID начального узла == 0) по массиву ID узлов ветки (cond []int)
Поиск начинается с первого узла (по значению cond[0]).
Если первый узел найден, то findSampleBrange вызывается рекурсивно для поиска следующего узла
и так далее, пока не будет найден удел последнего члена cond.
Если для текущего cond[level] узел в редеве не существует, то в дереве доращивается ветка по значениям cond []int.

Возвращает ID последней найденной ветки и ее уровень, если конечный узел не найден (если найден, то len(cond)-1).
Если вернула уровень меньше, чем максимальное число уровней ветки,
то ветка доращивается func addNodesToBrange

Ищет до узла Action включительно, а к этому узлу могут быть прикреплены несколько PARAMS
*/
func findEpisodicBrange(level int, cond []int, root *EpisodicTreeNode) (int,int) {

	// Обработка случая когда мы достигли конца дерева
	if cond == nil || len(cond) <= 0 || level >= len(cond) {
		return root.ID,len(cond)
	}

	// Поиск узла с ID из списка cond в дочерних узлах текущего узла
	for _, child := range root.Children {
		if isEquivalentCondition(level, &child,cond) {
			// Если узел найден, продолжим рекурсивно искать в нем далее
			id,lev := findEpisodicBrange(level+1, cond, &child)
			if id == 0 {
				/*Eсли условие cond []int содержит значение 0 для какого-то уровня, то на этом уровне поиск заканчивается.
				if cond[lev]==0{// считается нормальный поиск по неполному условию

				}*/
				return child.ID,lev // поиск закончен на child.ID
				//return 0,lev // не найдено совпадение на данном уровне
			}

			 return id,lev
		}
		//  пусть перебирает дочки! return 0,level // не найдено совпадение на данном уровне
	}

	// не найден узел на данном уровне
	return 0,level
}
////////////////////////////////////////////////////
func isEquivalentCondition(level int, node *EpisodicTreeNode,cond []int)bool{
	// массив значений по уровням - в зависимости от числа уровней в ветке, прописывается вручную
	nArr:= []int{node.NodeAID,node.NodeSID,node.NodePID,node.Trigger,node.Action}

	for i := 0; i < len(cond) && i<=level; i++ {
		if cond[i] != nArr[i] {
			return false
		}
	}
	return true
}
/////////////////////////////////////////////////////////




/* загрузить записанное дерево
 */
func loadEpisodicTree() {
	strArr, _ := lib.ReadLines(lib.GetMainPathExeFile() + "/memory_psy/episodic_tree.txt")
	initEpisodicTree(strArr)

	str := lib.ReadFileContent(lib.GetMainPathExeFile() + "/memory_psy/episodic_history.txt")
	sArr := strings.Split(str, "|")
	var p int
	for n := 0; n < len(sArr); n++ {
		if (len(sArr[n]) == 0) {
			break
		}
		p, _ = strconv.Atoi(sArr[n])
		EpisodicHistoryArr = append(EpisodicHistoryArr, p)
	}

	// ОТЛАДКА, ТЕСТИРОВАНИЕ
	// saveNewEpisodic(67,38,107,1,3,3)
	//	saveNewEpisodic(200,39,100,1,3,3)

	// getEpisodesArrFromConditions(0,curStimulImageID,0)
	// getEpisodesArrFromConditions(0,1,0)
	// getEpisodesArrFromConditions(1,1,0)

	// тестирование getTargetEpisodicStrIdArr()
	if false {// здесь еще не активированные деревья
		detectedActiveLastNodID=67
		detectedActiveLastUnderstandingNodID=38
		detectedActiveLastProblemNodID=107
		getTargetEpisodicStrIdArr(1, 5, 20)
	}

return
}
func initEpisodicTree(strArr []string){
	lib.MapCheckBlock(MapGwardEpisodicTreeNodeFromID)
	EpisodicTreeNodeFromID=make(map[int]*EpisodicTreeNode)
	lib.MapFree(MapGwardEpisodicTreeNodeFromID)
	// определить нулевой узел
	EpisodicTreeNodeFromID[0]=&EpisodicTree// все по нулям по умолчанию

	cunt:=len(strArr)
	//просто проход по всем строкам файла подряд так что сначала идут дочки, потом - их родители
	for n := 0; n < cunt; n++ {
		if len(strArr[n])==0{
			continue
		}
		if len(strArr[n])<2{
			panic("Сбой загрузки дерева: ["+strconv.Itoa(n) + "] " + strArr[n])
			return
		}
		p:=strings.Split(strArr[n], "|")
		id,_:=strconv.Atoi(p[0])
		parentID,_:=strconv.Atoi(p[1])
		NodeAID,_:=strconv.Atoi(p[2])
		NodeSID,_:=strconv.Atoi(p[3])
		NodePID,_:=strconv.Atoi(p[4])
		Trigger,_:=strconv.Atoi(p[5])
		Action,_:=strconv.Atoi(p[6])

		// считывание PARAMS по 3 значения из последовательности любой длины (но всегда кратной 3)
		var params = make([][]int, 0)
		if Action>0 {
			parArr := p[7:]
			/*
				// Добавляем числа в группы по 2
				for i := 0; i < len(parArr); i += 2 {
					end := i + 2
					if end > len(parArr) {
						end = len(parArr)
					}
					group := make([]int, 0)
					for _, numStr := range parArr[i:end] {
						num, _ := strconv.Atoi(numStr)
						group = append(group, num)
					}
					if len(group)<3{
						//continue
						break
					}
					params = append(params, group)
				}
			*/
			var n = 0
			var group []int
			for i := 0; i < len(parArr); i++ {

				num, _ := strconv.Atoi(parArr[i])
				group = append(group, num)
				n++
				// здесь  чтобы отработало последнюю
				if n%2 == 0 {
					params = append(params, group)
					group = nil
				}
			}
		}
		// новый узел с каждой строкой из файла
		lib.MapCheckBlock(MapGwardEpisodicTreeNodeFromID)
		parent,ok:=EpisodicTreeNodeFromID[parentID]
		lib.MapFree(MapGwardEpisodicTreeNodeFromID)
		if ok {
			// false - не проверять наличие ветки т.к. еще нет веток, они только глузятся
			createEpisodicNodeTree(parent, id, NodeAID, NodeSID, NodePID, Trigger, Action, params)
		}
	}
	return
}
/////////////////////////////////////
func SaveEpisodicTree(){
	var out=""
	cnt:=len(EpisodicTree.Children)
	for n := 0; n < cnt; n++ { // чтобы записывалось по порядку родителей
		out+=getEpisodicTreeNode(&EpisodicTree.Children[n])
	}
	lib.WriteFileContent(lib.GetMainPathExeFile()+"/memory_psy/episodic_tree.txt",out)

	// запись стории эпизодов EpisodicHistoryArr []int
	out=""
	for n := 0; n < len(EpisodicHistoryArr); n++ {
		out+=strconv.Itoa(EpisodicHistoryArr[n]) + "|"
	}
	lib.WriteFileContent(lib.GetMainPathExeFile()+"/memory_psy/episodic_history.txt",out)

	return
}
func getEpisodicTreeNode(wt *EpisodicTreeNode)(string){
	var out=""
	//	if wt.ParentID>0 {
	out += strconv.Itoa(wt.ID) + "|"
	out += strconv.Itoa(wt.ParentID) + "|"
	out += strconv.Itoa(wt.NodeAID) + "|"
	out += strconv.Itoa(wt.NodeSID) + "|"
	out += strconv.Itoa(wt.NodePID) + "|"
	out += strconv.Itoa(wt.Trigger) + "|"
	out += strconv.Itoa(wt.Action) + "|"

	// записать массивы PARAMS
	parsArr:=getEpisodicPARAMS(wt.ID)
	for n := 0; n < len(parsArr); n++ {
		out += strconv.Itoa(parsArr[n][0]) + "|"
		out += strconv.Itoa(parsArr[n][1]) + "|"
	}

	out +="\r\n"
	//	}
	if wt.Children==nil{// конец
		return out
	}
	for n := 0; n < len(wt.Children); n++ {
		out+=getEpisodicTreeNode(&wt.Children[n])
	}
	return out
}
/////////////////////////////////////
